@@ -59,7 +59,7 @@ def lineno():
     """Returns the current line number in our program."""
     return inspect.currentframe().f_back.f_lineno
 
-version = __version__ = "2.1"
+version = __version__ = "2.3"
 
 tokens = [
     'NUMBER',
@@ -92,9 +92,10 @@ tokens = [
     'CHAR_LITERAL', 
     'STRING_LITERAL',
     'NEW_LINE',
+    'SQUOTE',
 ]
 
-t_ignore = " \r.?@'"
+t_ignore = " \r.?@\f"
 t_NUMBER = r'[0-9][0-9XxA-Fa-f]*'
 t_NAME = r'[<>A-Za-z_~][A-Za-z0-9_]*'
 t_OPEN_PAREN = r'\('
@@ -124,6 +125,7 @@ def t_COMMENT_SINGLELINE(t):
             doxygenCommentCache += t.value[:-1]
         else:
             doxygenCommentCache += t.value
+    t.lexer.lineno += len([a for a in t.value if a=="\n"])
 t_ASTERISK = r'\*'
 t_MINUS = r'\-'
 t_PLUS = r'\+'
@@ -131,6 +133,7 @@ t_DIVIDE = r'/[^/]'
 t_AMPERSTAND = r'&'
 t_EQUALS = r'='
 t_CHAR_LITERAL = "'.'"
+t_SQUOTE = "'"
 #found at http://wordaligned.org/articles/string-literals-and-regular-expressions
 #TODO: This does not work with the string "bla \" bla"
 t_STRING_LITERAL = r'"([^"\\]|\\.)*"'
@@ -144,6 +147,7 @@ def t_COMMENT_MULTILINE(t):
         #strip prefixing whitespace
         v = re.sub("\n[\s]+\*", "\n*", v)
         doxygenCommentCache += v
+    t.lexer.lineno += len([a for a in t.value if a=="\n"])
 def t_NEWLINE(t):
     r'\n+'
     t.lexer.lineno += len(t.value)
@@ -268,6 +272,33 @@ def is_property_namestack(nameStack):
     if not r and is_function_pointer_stack(nameStack): r = True
     return r
 
+def detect_lineno(s):
+    """Detect the line number for a given token string"""
+    try:
+        rtn = s.lineno()
+        if rtn != -1:
+            return rtn
+    except: pass
+    global curLine
+    return curLine 
+
+class TagStr(str):
+    """Wrapper for a string that allows us to store the line number associated with it"""
+    lineno_reg = {}
+    def __new__(cls,*args,**kw):
+        new_obj =  str.__new__(cls,*args)
+        if "lineno" in kw:
+            TagStr.lineno_reg[id(new_obj)] = kw["lineno"]
+        return new_obj
+    
+    def __del__(self):
+        try:
+            del TagStr.lineno_reg[id(self)]
+        except: pass
+    
+    def lineno(self):
+        return TagStr.lineno_reg.get(id(self), -1)
+
 class CppParseError(Exception): pass
     
 class CppClass(dict):
@@ -356,6 +387,7 @@ class CppClass(dict):
             self["doxygen"] = doxygenCommentCache
             doxygenCommentCache = ""
         self["name"] = nameStack[1]
+        self["line_number"] = detect_lineno(nameStack[0])
         
         #Handle template classes
         if len(nameStack) > 3 and nameStack[2].startswith("<"):
@@ -373,12 +405,11 @@ class CppClass(dict):
                     self["name"] = "".join(nameStack[1:i + 1])
                     break;
                 i += 1
-                
-        inheritList = []
-
-        if ":" in nameStack:
+        elif ":" in nameStack:
             self['name'] = nameStack[ nameStack.index(':') - 1 ]
 
+        inheritList = []
+
         if nameStack.count(':') == 1:
             nameStack = nameStack[nameStack.index(":") + 1:]
             while len(nameStack):
@@ -702,6 +733,7 @@ class CppMethod( _CppMethod ):
                 break        
 
         self.update( methinfo )
+        self["line_number"] = detect_lineno(nameStack[0])
 
         #Filter out initializer lists used in constructors
         try:
@@ -840,7 +872,8 @@ class CppVariable( _CppVariable ):
             doxygenCommentCache = ""
 
         debug_print( "Variable: %s"%nameStack )
-        
+
+        self["line_number"] = detect_lineno(nameStack[0])
         self["function_pointer"] = 0
 
         if (len(nameStack) < 2):    # +++
@@ -881,7 +914,7 @@ class CppVariable( _CppVariable ):
         CppVariable.Vars.append( self )        # save and resolve later
     
     def __repr__(self):
-        keys_white_list = ['constant','name','reference','type','static','pointer','desc']
+        keys_white_list = ['constant','name','reference','type','static','pointer','desc', 'line_number']
         cpy = dict((k,v) for (k,v) in list(self.items()) if k in keys_white_list)
         if "array_size" in self: cpy["array_size"] = self["array_size"]
         return "%s"%cpy
@@ -904,6 +937,9 @@ class _CppEnum(dict):
         for v in values:
             if 'value' in v:
                 a = v['value'].strip()
+                # Remove single quotes from single quoted chars (unless part of some expression
+                if len(a) == 3 and a[0] == "'" and a[2] == "'":
+                    a = v['value'] = a[1]
                 if a.lower().startswith("0x"):
                     try:
                         i = a = int(a , 16)
@@ -926,6 +962,9 @@ class _CppEnum(dict):
                     v['raw_value'] = v['value']
                 v['value'] = a
             else: v['value'] = i
+            try:
+                v['value'] = v['value'].replace(" < < ", " << ").replace(" >> ", " >> ")
+            except: pass
             i += 1
         return t
 
@@ -942,14 +981,20 @@ class CppEnum(_CppEnum):
                 }
     """
     def __init__(self, nameStack):
-        if len(nameStack) < 4 or "{" not in nameStack or "}" not in nameStack:
-            #Not enough stuff for an enum
-            return
         global doxygenCommentCache
         if len(doxygenCommentCache):
             self["doxygen"] = doxygenCommentCache
             doxygenCommentCache = ""
+        if len(nameStack) == 3 and nameStack[0] == "enum":
+            debug_print("Created enum as just name/value")
+            self["name"] = nameStack[1]
+            self["instances"]=[nameStack[2]]
+        if len(nameStack) < 4 or "{" not in nameStack or "}" not in nameStack:
+            #Not enough stuff for an enum
+            debug_print("Bad enum")
+            return
         valueList = []
+        self["line_number"] = detect_lineno(nameStack[0])
         #Figure out what values it has
         valueStack = nameStack[nameStack.index('{') + 1: nameStack.index('}')]
         while len(valueStack):
@@ -1001,6 +1046,8 @@ class CppStruct(dict):
         else: self['type'] = None
         self['fields'] = []
         self.Structs.append( self )
+        global curLine
+        self["line_number"] = curLine
 
 C99_NONSTANDARD = {
     'int8' : 'signed char',
@@ -1263,9 +1310,13 @@ class Resolver(object):
                                     elif enum['type'] is str:
                                         var['ctypes_type'] = 'ctypes.c_char_p'
                                         var['raw_type'] = 'char*'
-                                    if 'method' in var: var['enum'] = var['method']['path'] + '::' + enum['name']
-                                    else:    # class property
+                                    try:
+                                        if 'method' in var: var['enum'] = var['method']['path'] + '::' + enum['name']
+                                        else:    # class property
+                                            var['unresolved'] = True
+                                    except:
                                         var['unresolved'] = True
+                                        
                                     var['fundamental'] = True
 
                                 else: var['unresolved'] = True    # TODO klass._public_xxx
@@ -1385,6 +1436,26 @@ class Resolver(object):
                 var['unresolved'] = True
                 if 'method' in var: var['method']['unresolved_parameters'] = True
                 #var['raw_type'] = var['raw_type'][2:]
+        
+        # Take care of #defines and #pragmas etc
+        trace_print("Processing precomp_macro_buf: %s"%self._precomp_macro_buf)
+        for m in self._precomp_macro_buf:
+            macro = m.replace("<CppHeaderParser_newline_temp_replacement>\\n", "\n")
+            try:
+                if macro.lower().startswith("#define"):
+                    trace_print("Adding #define %s"%macro)
+                    self.defines.append(macro.split(" ", 1)[1].strip())
+                elif macro.lower().startswith("#pragma"):
+                    trace_print("Adding #pragma %s"%macro)
+                    self.pragmas.append(macro.split(" ", 1)[1].strip())
+                elif macro.lower().startswith("#include"):
+                    trace_print("Adding #include %s"%macro)
+                    self.includes.append(macro.split(" ", 1)[1].strip())
+                else:
+                    debug_print("Cant detect what to do with precomp macro '%s'"%macro)
+            except: pass
+        self._precomp_macro_buf = None
+        
 
     def concrete_typedef( self, key ):
         if key not in self.typedefs:
@@ -1545,10 +1616,8 @@ class _CppHeader( Resolver ):
 
         if '{' in stack:
             info['defined'] = True
-            self._method_body = self.braceDepth
+            self._method_body = self.braceDepth + 1
             trace_print( 'NEW METHOD WITH BODY', self.braceDepth )
-            #if not self.curClass and ':' in stack: #Method/function defined outside class braces
-            #    return None
         elif stack[-1] == ';':
             info['defined'] = False
             self._method_body = None    # not a great idea to be clearing here
@@ -1756,6 +1825,7 @@ class _CppHeader( Resolver ):
         debug_print("curAccessSpecifier changed/defaulted to %s"%self.curAccessSpecifier)
         if self.nameStack[0] == "union":
             newClass = CppUnion(self.nameStack)
+            self.anon_union_counter = [self.braceDepth, 2]
             trace_print( 'NEW UNION', newClass['name'] )
         else:
             newClass = CppClass(self.nameStack)
@@ -1785,7 +1855,7 @@ class _CppHeader( Resolver ):
             self.curClass = newClass["name"]
             self._classes_brace_level[ newClass['name'] ] = self.braceDepth
 
-        if not key.endswith("::") and len(key) != 0:
+        if not key.endswith("::") and not key.endswith(" ") and len(key) != 0:
             if key in self.classes:
                 trace_print( 'ERROR name collision:', key )
                 self.classes[key].show()
@@ -1850,6 +1920,11 @@ class CppHeader( _CppHeader ):
         self.classes = Resolver.CLASSES
         #Functions that are not part of a class
         self.functions = []
+        
+        self.pragmas = []
+        self.defines = []
+        self.includes = []
+        self._precomp_macro_buf = [] #for internal purposes, will end up filling out pragmras and defines at the end
 
         self.enums = []
         self.global_enums = {}
@@ -1861,6 +1936,8 @@ class CppHeader( _CppHeader ):
         debug_print("curAccessSpecifier changed/defaulted to %s"%self.curAccessSpecifier)
         self.initextra()
     
+        self.anon_union_counter = [-1, 0]
+    
         if (len(self.headerFileName)):
             fd = open(self.headerFileName)
             headerFileStr = "".join(fd.readlines())
@@ -1873,27 +1950,54 @@ class CppHeader( _CppHeader ):
         
         # Strip out template declarations
         headerFileStr = re.sub("template[\t ]*<[^>]*>", "", headerFileStr)
-        
-        # Strip out #defines
+
+        # Change multi line #defines and expressions to single lines maintaining line nubmers
         # Based from http://stackoverflow.com/questions/2424458/regular-expression-to-match-cs-multiline-preprocessor-statements
-        headerFileStr = re.sub(r'(?m)^#[Dd][Ee][Ff][Ii][Nn][Ee] (?:.*\\\r?\n)*.*$', "", headerFileStr)
+        matches = re.findall(r'(?m)^(?:.*\\\r?\n)+.*$', headerFileStr)
+        is_define = re.compile(r'[ \t\v]*#[Dd][Ee][Ff][Ii][Nn][Ee]')
+        for m in matches:
+            #Keep the newlines so that linecount doesnt break
+            num_newlines = len([a for a in m if a=="\n"])
+            if is_define.match(m):
+                new_m = m.replace("\n", "<CppHeaderParser_newline_temp_replacement>\\n")
+            else:
+                # Just expression taking up multiple lines, make it take 1 line for easier parsing
+                new_m = m.replace("\\\n", " ")
+            if (num_newlines > 1):
+                new_m += "\n"*(num_newlines)
+            headerFileStr = headerFileStr.replace(m, new_m)
         
         #Filter out Extern "C" statements.  These are order dependent
-        headerFileStr = re.sub(re.compile(r'extern[\t ]+"[Cc]"[\t \n\r]*{', re.DOTALL), "namespace __IGNORED_NAMESPACE__CppHeaderParser__ {", headerFileStr) #To be ignored later
+        matches = re.findall(re.compile(r'extern[\t ]+"[Cc]"[\t \n\r]*{', re.DOTALL), headerFileStr)
+        for m in matches:
+            #Keep the newlines so that linecount doesnt break
+            num_newlines = len([a for a in m if a=="\n"])
+            headerFileStr = headerFileStr.replace(m, "\n" * num_newlines)        
         headerFileStr = re.sub(r'extern[ ]+"[Cc]"[ ]*', "", headerFileStr)
-        
+                
         self.braceDepth = 0
+        lex.lex()
         lex.input(headerFileStr)
+        global curLine
+        global curChar
         curLine = 0
         curChar = 0
         try:
             while True:
                 tok = lex.token()
                 if not tok: break
+                if self.anon_union_counter[0] == self.braceDepth and self.anon_union_counter[1]:
+                    self.anon_union_counter[1] -= 1
+                tok.value = TagStr(tok.value, lineno=tok.lineno)
+                #debug_print("TOK: %s"%tok)
                 if tok.type == 'NAME' and tok.value in self.IGNORE_NAMES: continue
-                if tok.type not in ('PRECOMP_MACRO', 'PRECOMP_MACRO_CONT'): self.stack.append( tok.value )
+                self.stack.append( tok.value )
                 curLine = tok.lineno
                 curChar = tok.lexpos
+                if (tok.type in ('PRECOMP_MACRO', 'PRECOMP_MACRO_CONT')):
+                    debug_print("PRECOMP: %s"%tok)
+                    self._precomp_macro_buf.append(tok.value)
+                    continue
                 if (tok.type == 'OPEN_BRACE'):
                     if len(self.nameStack) >= 2 and is_namespace(self.nameStack):    # namespace {} with no name used in boost, this sets default?
                         if self.nameStack[1] == "__IGNORED_NAMESPACE__CppHeaderParser__":#Used in filtering extern "C"
@@ -1938,7 +2042,7 @@ class CppHeader( _CppHeader ):
                         trace_print( 'END OF STRUCT DEF' )
                         self.curStruct = None
 
-                    if self._method_body and self.braceDepth <= self._method_body:
+                    if self._method_body and (self.braceDepth + 1) <= self._method_body:
                         self._method_body = None; self.stack = []; self.nameStack = []; trace_print( 'FORCE CLEAR METHBODY' )
                 
                 if (tok.type == 'OPEN_PAREN'):
@@ -1964,6 +2068,7 @@ class CppHeader( _CppHeader ):
                     self.nameStack.append(tok.value)
                 elif (tok.type == 'EXCLAMATION'):
                     self.nameStack.append(tok.value)
+                elif (tok.type == 'SQUOTE'): pass
                 elif (tok.type == 'NUMBER'):
                     self.nameStack.append(tok.value)
                 elif (tok.type == 'MINUS'):
@@ -1972,7 +2077,7 @@ class CppHeader( _CppHeader ):
                     self.nameStack.append(tok.value)
                 elif (tok.type == 'STRING_LITERAL'):
                     self.nameStack.append(tok.value)
-                elif (tok.type == 'NAME' or tok.type == 'AMPERSTAND' or tok.type == 'ASTERISK'):
+                elif (tok.type == 'NAME' or tok.type == 'AMPERSTAND' or tok.type == 'ASTERISK' or tok.type == 'CHAR_LITERAL'):
                     if tok.value in ignoreSymbols:
                         debug_print("Ignore symbol %s"%tok.value)
                     elif (tok.value == 'class'):
@@ -1987,6 +2092,8 @@ class CppHeader( _CppHeader ):
                         self.stack = []
                     else:
                         self.nameStack.append(tok.value)
+                        if self.anon_union_counter[0] == self.braceDepth:
+                            self.anon_union_counter = [-1, 0]
                 elif (tok.type == 'COLON'):
                     #Dont want colon to be first in stack
                     if len(self.nameStack) == 0:
@@ -2005,6 +2112,22 @@ class CppHeader( _CppHeader ):
                     self.accessSpecifierScratch = []
 
                 elif (tok.type == 'SEMI_COLON'):
+                    if self.anon_union_counter[0] == self.braceDepth and self.anon_union_counter[1]:
+                        debug_print("Creating anonymous union")
+                        #Force the processing of an anonymous union
+                        saved_namestack = self.nameStack[:] 
+                        saved_stack = self.stack[:]
+                        self.nameStack = [""]
+                        self.stack = self.nameStack + [";"]
+                        self.nameStack = self.nameStack[0:1]
+                        debug_print("pre eval anon stack")
+                        self.evaluate_stack( tok.type )
+                        debug_print("post eval anon stack")
+                        self.nameStack = saved_namestack
+                        self.stack = saved_stack
+                        self.anon_union_counter = [-1, 0];
+                    
+                    
                     if (self.braceDepth < 10): self.evaluate_stack( tok.type )
                     if not self.stack: continue
                     if self.stack[0]=='typedef' and ( '{' not in self.stack or '}' in self.stack ): self.stack = []; trace_print( "REAL CLEAR")
@@ -2016,6 +2139,8 @@ class CppHeader( _CppHeader ):
                                 % (self.headerFileName, tok.lineno, tok.value, " ".join(self.nameStack)))
 
         self.finalize()
+        global parseHistory
+        parseHistory = []
 
     def evaluate_stack(self, token=None):
         """Evaluates the current name stack"""
@@ -2063,7 +2188,7 @@ class CppHeader( _CppHeader ):
             debug_print( "trace" )
             self.evaluate_enum_stack()
 
-        elif self._method_body and self.braceDepth > self._method_body: trace_print( 'INSIDE METHOD DEF' )
+        elif self._method_body and (self.braceDepth + 1) > self._method_body: trace_print( 'INSIDE METHOD DEF' )
         elif is_method_namestack(self.stack) and not self.curStruct and '(' in self.nameStack:
             debug_print( "trace" )
             if self.braceDepth > 0:
@@ -2117,6 +2242,7 @@ class CppHeader( _CppHeader ):
 
     def evaluate_enum_stack(self):
         """Create an Enum out of the name stack"""
+        debug_print( "evaluating enum" )
         newEnum = CppEnum(self.nameStack)
         if len(list(newEnum.keys())):
             if len(self.curClass):
@@ -2148,4 +2274,8 @@ class CppHeader( _CppHeader ):
             rtn += "// functions\n"
             for f in self.functions:
                 rtn += "%s\n"%f
+        if self.enums:
+            rtn += "// enums\n"
+            for f in self.enums:
+                rtn += "%s\n"%f
         return rtn
