@@ -1,6 +1,9 @@
 # CppHeaderParser package
 # Author: Jashua Cloutier (contact via sourceforge username:senexcanis)
+import sys
+if sys.version_info[0] == 2:
+    from CppHeaderParser import *
+else:
+    from CppHeaderParser3 import *
 
-from CppHeaderParser import *
-
-#__all__ = ['CppHeaderParser']
\ No newline at end of file
+#__all__ = ['CppHeaderParser']
