@@ -52,4 +52,12 @@ print cppHeader.classes["AlphaClass"]["properties"]["public"][0]["type"]
 
 print "\nFree functions are:"
 for func in cppHeader.functions:
-    print " %s"%func["name"]
\ No newline at end of file
+    print " %s"%func["name"]
+
+print "\n#includes are:"
+for incl in cppHeader.includes:
+    print " %s"%incl
+
+print "\n#defines are:"
+for define in cppHeader.defines:
+    print " %s"%define
\ No newline at end of file
